@@ -6,7 +6,7 @@
  *	License as published by the Free Software Foundation version 2.1
  *	of the License.
  *
- * Copyright (c) 2003-2008 Thomas Graf <tgraf@suug.ch>
+ * Copyright (c) 2003-2012 Thomas Graf <tgraf@suug.ch>
  */
 
 #ifndef NETLINK_LOCAL_TYPES_H_
@@ -17,6 +17,7 @@
 #include <netlink/route/qdisc.h>
 #include <netlink/route/rtnl.h>
 #include <netlink/route/route.h>
+#include <netlink/route/tc-api.h>
 
 #define NL_SOCK_BUFSIZE_SET	(1<<0)
 #define NL_SOCK_PASSCRED	(1<<1)
@@ -91,7 +92,7 @@ struct nl_cache_mngr
 	int			cm_protocol;
 	int			cm_flags;
 	int			cm_nassocs;
-	struct nl_sock *	cm_handle;
+	struct nl_sock *	cm_sock;
 	struct nl_cache_assoc *	cm_assocs;
 };
 
@@ -101,11 +102,6 @@ struct nl_parser_param;
 
 #define NL_OBJ_MARK		1
 
-struct nl_object
-{
-	NLHDR_COMMON
-};
-
 struct nl_data
 {
 	size_t			d_size;
@@ -150,29 +146,32 @@ struct rtnl_link
 {
 	NLHDR_COMMON
 
-	char		l_name[IFNAMSIZ];
-
-	uint32_t	l_family;
-	uint32_t	l_arptype;
-	uint32_t	l_index;
-	uint32_t	l_flags;
-	uint32_t	l_change;
-	uint32_t 	l_mtu;
-	uint32_t	l_link;
-	uint32_t	l_txqlen;
-	uint32_t	l_weight;
-	uint32_t	l_master;
-	struct nl_addr *l_addr;	
-	struct nl_addr *l_bcast;
-	char		l_qdisc[IFQDISCSIZ];
-	struct rtnl_link_map l_map;
-	uint64_t	l_stats[RTNL_LINK_STATS_MAX+1];
-	uint32_t	l_flag_mask;
-	uint8_t		l_operstate;
-	uint8_t		l_linkmode;
+	char				l_name[IFNAMSIZ];
+	uint32_t			l_family;
+	uint32_t			l_arptype;
+	uint32_t			l_index;
+	uint32_t			l_flags;
+	uint32_t			l_change;
+	uint32_t 			l_mtu;
+	uint32_t			l_link;
+	uint32_t			l_txqlen;
+	uint32_t			l_weight;
+	uint32_t			l_master;
+	struct nl_addr *		l_addr;	
+	struct nl_addr *		l_bcast;
+	char				l_qdisc[IFQDISCSIZ];
+	struct rtnl_link_map		l_map;
+	uint64_t			l_stats[RTNL_LINK_STATS_MAX+1];
+	uint32_t			l_flag_mask;
+	uint32_t			l_num_vf;
+	uint8_t				l_operstate;
+	uint8_t				l_linkmode;
 	/* 2 byte hole */
-	struct rtnl_link_info_ops *l_info_ops;
-	void *		l_info;
+	char *				l_info_kind;
+	struct rtnl_link_info_ops *	l_info_ops;
+	void *				l_af_data[AF_MAX];
+	void *				l_info;
+	char *				l_ifalias;
 };
 
 struct rtnl_ncacheinfo
@@ -236,6 +235,7 @@ struct rtnl_addr
 	
 	char a_label[IFNAMSIZ];
 	uint32_t a_flag_mask;
+	struct rtnl_link *a_link;
 };
 
 struct rtnl_nexthop
@@ -281,20 +281,21 @@ struct rtnl_route
 struct rtnl_rule
 {
 	NLHDR_COMMON
-
-	uint64_t	r_mark;
-	uint32_t	r_prio;
-	uint32_t	r_realms;
-	uint32_t	r_table;
-	uint8_t		r_dsfield;
-	uint8_t		r_type;
 	uint8_t		r_family;
-	uint8_t		r_src_len;
-	uint8_t		r_dst_len;
-	char		r_iif[IFNAMSIZ];
+	uint8_t		r_action;
+	uint8_t		r_dsfield; /* ipv4 only */
+	uint8_t		r_unused;
+	uint32_t	r_table;
+	uint32_t	r_flags;
+	uint32_t	r_prio;
+	uint32_t	r_mark;
+	uint32_t	r_mask;
+	uint32_t	r_goto;
+	uint32_t	r_flow; /* ipv4 only */
 	struct nl_addr *r_src;
 	struct nl_addr *r_dst;
-	struct nl_addr *r_srcmap;
+	char		r_iifname[IFNAMSIZ];
+	char		r_oifname[IFNAMSIZ];
 };
 
 struct rtnl_neightbl_parms
@@ -414,8 +415,8 @@ struct rtnl_neightbl
 struct rtnl_ratespec
 {
 	uint8_t			rs_cell_log;
-	uint16_t		rs_feature;
-	uint16_t		rs_addend;
+	uint16_t		rs_overhead;
+	int16_t			rs_cell_align;
 	uint16_t		rs_mpu;
 	uint32_t		rs_rate;
 };
@@ -443,43 +444,46 @@ struct rtnl_tstats
 
 #define TCKINDSIZ	32
 
-#define NL_TCA_GENERIC(pre)				\
+#define NL_TC_GENERIC(pre)				\
 	NLHDR_COMMON					\
 	uint32_t		pre ##_family;		\
 	uint32_t		pre ##_ifindex;		\
 	uint32_t		pre ##_handle;		\
 	uint32_t		pre ##_parent;		\
 	uint32_t		pre ##_info;		\
+	uint32_t		pre ##_mtu;		\
+	uint32_t		pre ##_mpu;		\
+	uint32_t		pre ##_overhead;	\
+	uint32_t		pre ##_linktype;	\
 	char			pre ##_kind[TCKINDSIZ];	\
 	struct nl_data *	pre ##_opts;		\
 	uint64_t		pre ##_stats[RTNL_TC_STATS_MAX+1]; \
 	struct nl_data *	pre ##_xstats;		\
 	struct nl_data *	pre ##_subdata;		\
+	struct rtnl_link *	pre ##_link;		\
+	struct rtnl_tc_ops *	pre ##_ops;		\
+	enum rtnl_tc_type	pre ##_type
 
-
-struct rtnl_tca
+struct rtnl_tc
 {
-	NL_TCA_GENERIC(tc);
+	NL_TC_GENERIC(tc);
 };
 
 struct rtnl_qdisc
 {
-	NL_TCA_GENERIC(q);
-	struct rtnl_qdisc_ops	*q_ops;
+	NL_TC_GENERIC(q);
 };
 
 struct rtnl_class
 {
-	NL_TCA_GENERIC(c);
-	struct rtnl_class_ops	*c_ops;
+	NL_TC_GENERIC(c);
 };
 
 struct rtnl_cls
 {
-	NL_TCA_GENERIC(c);
+	NL_TC_GENERIC(c);
 	uint16_t		c_prio;
 	uint16_t		c_protocol;
-	struct rtnl_cls_ops	*c_ops;
 };
 
 struct rtnl_u32
@@ -516,12 +520,14 @@ struct rtnl_ematch
 	uint16_t		e_id;
 	uint16_t		e_kind;
 	uint16_t		e_flags;
+	uint16_t		e_index;
+	size_t			e_datalen;
 
 	struct nl_list_head	e_childs;
 	struct nl_list_head	e_list;
 	struct rtnl_ematch_ops *e_ops;
 
-	char			e_data[0];
+	void *			e_data;
 };
 
 struct rtnl_ematch_tree
@@ -562,7 +568,6 @@ struct rtnl_prio
 struct rtnl_tbf
 {
 	uint32_t		qt_limit;
-	uint32_t		qt_mpu;
 	struct rtnl_ratespec	qt_rate;
 	uint32_t		qt_rate_bucket;
 	uint32_t		qt_rate_txtime;
@@ -627,20 +632,19 @@ struct rtnl_htb_qdisc
 	uint32_t		qh_rate2quantum;
 	uint32_t		qh_defcls;
 	uint32_t		qh_mask;
+	uint32_t		qh_direct_pkts;
 };
 
 struct rtnl_htb_class
 {
 	uint32_t		ch_prio;
-	uint32_t		ch_mtu;
 	struct rtnl_ratespec	ch_rate;
 	struct rtnl_ratespec	ch_ceil;
 	uint32_t		ch_rbuffer;
 	uint32_t		ch_cbuffer;
 	uint32_t		ch_quantum;
-	uint8_t			ch_overhead;
-	uint8_t			ch_mpu;
 	uint32_t		ch_mask;
+	uint32_t		ch_level;
 };
 
 struct rtnl_cbq
@@ -665,6 +669,12 @@ struct rtnl_red
 	uint32_t	qr_mask;
 };
 
+struct rtnl_plug
+{
+	int             action;
+	uint32_t        limit;
+};
+
 struct flnl_request
 {
 	NLHDR_COMMON
@@ -834,4 +844,10 @@ struct nfnl_queue_msg {
 	uint32_t		queue_msg_verdict;
 };
 
+struct ematch_quoted {
+	char *	data;
+	size_t	len;
+	int	index;
+};
+
 #endif
